import openai
from openai.types.chat import ChatCompletionToolParam, ChatCompletionNamedToolChoiceParam
from openai import RateLimitError, AuthenticationError, OpenAI
from rich import print
from tenacity import retry, wait_random_exponential, stop_after_attempt, retry_if_not_exception_type


@retry(
    retry=retry_if_not_exception_type((RateLimitError, AuthenticationError)),
    wait=wait_random_exponential(multiplier=1, max=40),
    stop=stop_after_attempt(3)
)
def chat_completion_request(
    openai_client: OpenAI,
    messages, functions=None, function_call=None, model="gpt-4o-mini"
):
    call_data = {"model": model, "messages": messages}

    if functions is not None:
        call_data["tools"] = [
            ChatCompletionToolParam(function=f, type="function") for f in functions
        ]

    if function_call is not None:
        print(f"Calling function {function_call}")
        call_data["tool_choice"] = ChatCompletionNamedToolChoiceParam(
            function=function_call,
            type='function',
        )
    try:
        return openai_client.chat.completions.create(**call_data)
    except Exception as e:
        print("Unable to generate ChatCompletion response")
        print(f"Exception: {e}")
        raise e



INITIAL_NON_INTERACTIVE_PROMPT = "How can I help you explore your database?"
INITIAL_INTERACTIVE_PROMPT = "How can I help you explore your database?"
FOLLOW_UP_PROMPT = "anything else I can help you with?"



import httpx


class WikiDataQueryTool:
    """
    For example to select the largest cities in the world that have a female mayor, you can use the following query:

    SELECT ?cityLabel ?mayorLabel WHERE { ?city wdt:P31 wd:Q515. ?city wdt:P6 ?mayor. ?mayor wdt:P21 wd:Q6581072. ?city wdt:P1082 ?population. SERVICE wikibase:label { bd:serviceParam
    wikibase:language 'en'. } } ORDER BY DESC(?population) LIMIT 10

    Or to get billionaires:

        SELECT ?locationLabel ?item ?itemLabel (MAX(?billion) as ?billions)
        WHERE
        {
          ?item wdt:P2218 ?worth.
          ?item wdt:P19 ?location .

          FILTER(?worth>1000000000).
          BIND(?worth/1000000000 AS ?billion).
          SERVICE wikibase:label { bd:serviceParam wikibase:language "en,de". }
        }
        GROUP BY ?locationLabel ?item ?itemLabel
        ORDER BY DESC(?billions)
        LIMIT 10

    For example to answer "How many Hospitals are there located in Beijing", you can use the following query:

        SELECT (COUNT(?hospital) AS ?count) WHERE { ?hospital wdt:P31 wd:Q16917 . ?hospital wdt:P131 wd:Q956 . SERVICE wikibase:label { bd:serviceParam wikibase:language '[AUTO_LANGUAGE],en'. } }
        LIMIT 10

    Retrieve the names of the Star Wars films:

        SELECT ?item  ?itemLabel
        WHERE
        {
          ?item wdt:P179 wd:Q22092344.
        SERVICE wikibase:label { bd:serviceParam wikibase:language "[AUTO_LANGUAGE]". }
        }
    """

    name = "wikidata"
    description = """Useful for when you need specific data from Wikidata.
    Input to this tool is a single correct SPARQL statement for Wikidata. Limit all requests to 10 or fewer rows. 
    
    Output is the raw response in json. If the query is not correct, an error message will be returned. 
    If an error is returned, you may rewrite the query and try again. If you are unsure about the response
    you can try rewrite the query and try again. Prefer local data before using this tool.
    """
    base_url: str = "https://query.wikidata.org/sparql"
    httpx_client: httpx.AsyncClient = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.httpx_client = httpx.AsyncClient()

    def _run(self, query: str) -> str:
        r = httpx.get(
            self.base_url, params={"format": "json", "query": query}, timeout=60
        )
        data = r.text  # no point parsing the json
        return data

    async def _arun(self, query: str) -> str:
        r = await self.httpx_client.get(
            self.base_url, params={"format": "json", "query": query}, timeout=60
        )
        data = r.text
        return data



import json
import textwrap
from typing import Callable, List
from pydantic import BaseModel
from openai.types.chat import (
    ChatCompletionSystemMessageParam,
    ChatCompletionToolMessageParam,
    ChatCompletionMessageParam, ChatCompletionMessageToolCallParam, ChatCompletionAssistantMessageParam,
)
from openai.types.chat.chat_completion_message_tool_call import Function
from openai import RateLimitError, OpenAI
from rich import print

from qabot.config import AgentModelConfig
from qabot.formatting import format_robot, format_duck, format_user
from qabot.functions import get_function_specifications
from qabot.functions.data_loader import import_into_duckdb_from_files
from qabot.functions.describe_duckdb_table import describe_table_or_view
from qabot.functions.duckdb_query import run_sql_catch_error
from qabot.functions.wikidata import WikiDataQueryTool
from qabot.llm import chat_completion_request
from qabot.prompts.system import system_prompt, research_prompt


class Agent:
    """
    An Agent is a wrapper around a Language Model that can be used to interact with the LLM.
    Our agent also wraps the functions that can be called by the LLM.
    """

    def __init__(
            self,
            database_engine=None,
            models: AgentModelConfig = None,
            prompt_context: str = None,
            allow_wikidata: bool = False,
            terminate_session_callback: Callable = None,
            clarification_callback: Callable[[str], str] | None = None,
            verbose=False,
            max_iterations: int = 20,
            openai_client: OpenAI = None,
    ):
        """
        Create a new Agent.
        """
        self.max_iterations = max_iterations
        self.model_name = models.default_model_name
        self.planning_model_name = models.planning_model_name
        self.db = database_engine
        self.verbose = verbose
        if verbose:
            print(
                format_robot(
                    f"Default model: {models.default_model_name}, Thinking model: {models.planning_model_name}. Max LLM/function iterations before answer {max_iterations}"
                )
            )
        self.functions = {
            "terminate_session": terminate_session_callback,
            "clarify": clarification_callback,
            "wikidata": lambda query: WikiDataQueryTool()._run(query),
            "execute_sql": lambda query: run_sql_catch_error(database_engine, query),
            "show_tables": lambda: run_sql_catch_error(database_engine,
                                                       "select table_catalog, table_schema, table_name from system.information_schema.tables where table_schema != 'information_schema';"),
            "describe_table": lambda table, **kwargs: describe_table_or_view(
                database_engine, table, **kwargs
            ),
            "research": self.research_call,
            "load_data": lambda files: "Imported with SQL:\n"
                                       + str(import_into_duckdb_from_files(database_engine, files)[1]),
        }
        self.function_specifications = get_function_specifications(allow_wikidata, allow_research=True)

        if clarification_callback is not None:
            self.function_specifications.append(
                {
                    "name": "clarify",
                    "description": textwrap.dedent(
                        """Useful for when you need to ask the user a question to clarify their request.
                        Input to this tool is a single question for the user. Output is the user's response.
                        """
                    ),
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "clarification": {
                                "type": "string",
                                "description": "A question or prompt for the user",
                            },
                        },
                    },
                }
            )

        if terminate_session_callback is not None:
            self.function_specifications.append(
                {
                    "name": "terminate_session",
                    "description": textwrap.dedent(
                        """Indicate that the user has requested the session to be terminated.
                        Prefer to clarify unclear requests or requests you cannot address rather than terminating a session immediately.
                        """
                    ),
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "message": {
                                "type": "string",
                                "description": "The final short message - usually saying goodbye",
                            },
                        },
                    },
                }
            )

        messages: List[ChatCompletionMessageParam] = [
            ChatCompletionSystemMessageParam(
                **{"role": "system", "content": system_prompt}
            ),

            # Force the assistant to get the current tables
            ChatCompletionAssistantMessageParam(
                role="assistant",
                content='',
                tool_calls=[
                    ChatCompletionMessageToolCallParam(
                        type="function",
                        id="show_tables",
                        function=Function(name="show_tables", arguments="{}"),
                    )
                ],
            ),
        ]

        messages.append(
            ChatCompletionToolMessageParam(
                **{
                    "role": "tool",
                    "tool_call_id": "show_tables",
                    "content": execute_function_call(messages[-1]['tool_calls'][0]['function'], self.functions),
                }
            )
        )
        if prompt_context is not None:
            messages.append({"role": "user", "content": prompt_context})

        self.messages = messages

        self.openai_client = openai_client or OpenAI()

    def __call__(self, user_input):
        """
        Pass new input from the user to the LLM and return the response.
        """
        return self.run(user_input)

    def run(self, user_input):
        """
        Run the LLM/function execution loop and return the final response.
        """
        self.messages.append({"role": "user", "content": user_input})

        for _ in range(self.max_iterations):
            is_final_answer, result = self.llm_step()

            if is_final_answer:
                return json.loads(result)

        # If we get here, we've hit the max number of iterations
        # Let's ask the LLM to summarize the errors/answer as best it can
        self.messages.extend(
            [
                {
                    "role": "system",
                    "content": "Maximum iterations reached. Summarize what you were doing and attempt to answer the users question",
                }
            ]
        )

        _, result = self.llm_step(forced_function_call={"name": "answer"})
        return result

    def llm_step(self, forced_function_call=None):
        is_final_answer = False
        function_call_results = None

        chat_response = chat_completion_request(
            self.openai_client,
            self.messages,
            functions=self.function_specifications,
            model=self.model_name,
            function_call=forced_function_call,
        )

        choice = chat_response.choices[0]
        message = choice.message
        self.messages.append(message)
        if hasattr(message, "tool_calls") and message.tool_calls:
            for tool_call in message.tool_calls:
                function_name = tool_call.function.name
                call_id = tool_call.id

                function_call_results = execute_function_call(
                    tool_call.function, self.functions, self.verbose
                )

                # Inject a response message for the function call
                self.messages.append(
                    ChatCompletionToolMessageParam(
                        content=function_call_results, role="tool", tool_call_id=call_id
                    )
                )

                is_final_answer = function_name == "answer"

                if self.verbose:
                    format_response = (
                        format_user if function_name == "clarification" else format_duck
                    )
                    print(format_response(function_call_results))
        if message.content is not None and self.verbose:
            print(format_robot(message.content))
        return is_final_answer, function_call_results

    def research_call(self, query):
        print("Research Time")
        # Now we use the planning LLM model
        chat_response = chat_completion_request(
            self.openai_client,
            messages=[
                         ChatCompletionSystemMessageParam(
                             **{"role": "system", "content": research_prompt}
                         ),
                     ] +
                     [{"role": "system", "content": str(m)} for m in self.messages[-10:]]
                     +
                     [
                         ChatCompletionSystemMessageParam(
                             **{"role": "system", "content": "Question follows:"}
                         ),
                         {"role": "user", "content": query}
                     ],
            model=self.planning_model_name
        )

        choice = chat_response.choices[0]
        message = choice.message

        if message.content is not None and self.verbose:
            print(format_robot(message.content))

        return message.content


def create_agent_executor(**kwargs):
    return Agent(**kwargs)


def execute_function_call(function, functions, verbose=False):
    function_name = function.name
    try:
        kwargs = json.loads(function.arguments)
    except json.decoder.JSONDecodeError:
        return "Error: function arguments were not valid JSON"

    if function_name in functions:
        f = functions[function_name]
        if verbose:
            if kwargs:
                print(format_robot(function_name), kwargs)
            else:
                print(format_robot(function_name))
        try:
            return f(**kwargs)
        except SystemExit as e:
            raise SystemExit(e)
        except Exception as e:
            results = f"Error: Calling function {function_name} raised an exception.\n\n{str(e)}"
    elif function_name == "answer":
        return json.dumps(kwargs)

    else:
        return f"Error: function {function_name} does not exist"

    return results



from typing import Optional

from qabot.agent import Agent, AgentModelConfig
from qabot.config import Settings
from qabot.functions.data_loader import create_duckdb, import_into_duckdb_from_files


def ask_wikidata(query: str, model_name=None, verbose=False):
    model_config = Settings().agent_model
    if model_name is not None:
        model_config.default_model_name = model_name

    agent = Agent(allow_wikidata=True, models=model_config, verbose=verbose)
    result = agent(query)
    return result["summary"]


def ask_file(query: str, filename: Optional[str], model_name=None, verbose=False):
    engine = create_duckdb()
    database_engine, executed_sql = import_into_duckdb_from_files(engine, [filename])
    model_config = Settings().agent_model
    if model_name is not None:
        model_config.default_model_name = model_name
    agent = Agent(database_engine=database_engine, models=model_config, verbose=verbose)
    result = agent(query)
    return result["summary"]


def ask_database(query: str, uri: str, model_name=None, context=None, verbose=False):
    engine = create_duckdb()
    model_config = Settings().agent_model
    if model_name is not None:
        model_config.default_model_name = model_name
    database_engine, executed_sql = import_into_duckdb_from_files(engine, [uri])
    agent = Agent(database_engine=database_engine, models=model_config, prompt_context=context, verbose=verbose)
    result = agent(query)
    return result["summary"]



import os
import httpx
from datetime import datetime, timedelta
import appdirs


def get_cache_dir(app_name: str) -> str:
    return appdirs.user_cache_dir(app_name)


def download_and_cache(url: str, cache_duration: timedelta = timedelta(days=30)):
    filename = url.split('/')[-1]
    cache_dir = get_cache_dir("qabot")
    file_path = os.path.join(cache_dir, filename)

    # Check if file exists and is within the cache duration
    if os.path.exists(file_path):
        file_mod_time = datetime.fromtimestamp(os.path.getmtime(file_path))
        if datetime.now() - file_mod_time < cache_duration:
            return file_path

    # Download and save the file
    response = httpx.get(url)
    if response.status_code == 200:
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, 'wb') as file:
            file.write(response.content)
    else:
        raise Exception("Failed to download the file.") from response.raise_for_status()

    return file_path


if __name__ == '__main__':
    doc_url = "https://duckdb.org/duckdb-docs.pdf"
    download_and_cache(doc_url)



import logging

from qabot.functions.duckdb_query import run_sql_catch_error


def describe_table_or_view(database, table: str , schema=None, catalog=None):
    """
    Show the column names and types of a local database table or view.

    Note if the catalog is not default we don't compute the size of the table.
    """
    logging.debug(f"describe_table_or_view({table}, {schema}, {catalog})")

    # Identify the catalog, schema and table - if not provided
    if catalog is None:
        # Search the system.information_schema.tables for the table
        schema_clause = f" and table_schema='{schema}'" if schema is not None else ""
        table_catalog_query = f"select table_catalog from system.information_schema.tables where table_name='{table}'{schema_clause} limit 1;"
        catalog_row = database.sql(table_catalog_query).fetchone()
        if not catalog_row:
            return f"Table {table} not found in any catalog"
        catalog = catalog_row[0]

    if schema is None:
        schema_query = f"select table_schema from system.information_schema.tables where table_name='{table}' and table_catalog='{catalog}' limit 1;"
        schema_row = database.sql(schema_query).fetchone()
        if schema_row is None:
            return f"Table {table} not found in catalog {catalog}"
        schema = schema_row[0]

    fully_qualified_table = f"{catalog}.{schema}.{table}"
    logging.debug(fully_qualified_table)
    # If the catalog is external, we avoid computing the size of the table
    if catalog == "memory":
        table_count_rows_query = f"select count(*) from {fully_qualified_table};"
        table_size = table_count_rows_query + '\n' + str(database.sql(table_count_rows_query).fetchone()[0])
    else:
        table_size = ""

    table_columns_and_types_query = f"select column_name, data_type from system.information_schema.columns where table_name='{table}' and table_catalog='{catalog}' and table_schema='{schema}';"

    # Get the names of the first 20 columns of the table
    column_name_query = f"select column_name from system.information_schema.columns where table_name='{table}' and table_catalog='{catalog}' and table_schema='{schema}' limit 20;"
    column_name_results = database.sql(column_name_query).fetchall()
    column_names = [f'"{row[0]}"' for row in column_name_results]
    joined_names = ", ".join(column_names)
    table_first_rows_query = f"select {joined_names} from {fully_qualified_table} limit 5;"

    table_description = run_sql_catch_error(database, table_columns_and_types_query)

    table_preview = run_sql_catch_error(database, table_first_rows_query)[:4000]
    return f"{table}\n{table_description}\n\n{table_size}\n{table_first_rows_query}\n{table_preview}"



from typing import List
from pydantic import AnyUrl, BaseModel, model_validator
from pydantic_settings import BaseSettings


class AgentModelConfig(BaseModel):
    default_model_name: str = "gpt-4o-mini"
    planning_model_name: str = 'o3-mini'


class Settings(BaseSettings):
    OPENAI_BASE_URL: str | None = None
    OPENAI_API_KEY: str

    QABOT_DATABASE_URI: str | None = None
    QABOT_CACHE_DATABASE_URI: AnyUrl = "duckdb:///:memory:"
    QABOT_MODEL_NAME: str = "gpt-4o-mini"
    QABOT_PLANNING_MODEL_NAME: str = "o3-mini"
    QABOT_TABLES: List[str] | None = None
    QABOT_ENABLE_WIKIDATA: bool = True
    QABOT_ENABLE_HUMAN_CLARIFICATION: bool = True

    agent_model: AgentModelConfig = AgentModelConfig()

    @model_validator(mode="before")
    def combine_agent_model(cls, values):
        # Build nested config from env values before other validation occurs
        values["agent_model"] = {
            "default_model_name": values.get("QABOT_MODEL_NAME", "gpt-4o-mini"),
            "planning_model_name": values.get("QABOT_PLANNING_MODEL_NAME", "o3-mini"),
        }
        return values

    class Config:
        env_prefix = ""  # if you want to avoid prefixes



from rich import print

role_to_color = {
    "system": "red",
    "user": "green",
    "assistant": "blue",
    "function": "magenta",
}

ROBOT_COLOR = "blue"
DUCK_COLOR = "magenta"
QUERY_COLOR = "cyan"

ROCKET_EMOJI = "[red] 🚀"
DUCK_EMOJI = f"[{DUCK_COLOR}] 🦆"
ROBOT_EMOJI = f"[{ROBOT_COLOR}] 🤖"
USER_EMOJI = "[green] 🧑"


def format_query(msg):
    return f"[{QUERY_COLOR}]{msg}[/{QUERY_COLOR}]"


def format_robot(msg):
    return f"{ROBOT_EMOJI} {msg}[/]"


def format_user(msg):
    return f"{USER_EMOJI} {msg}[/]"


def format_duck(msg):
    return f"{DUCK_EMOJI} {msg}[/]"


def format_rocket(msg):
    return f"{ROCKET_EMOJI} {msg}[/]"


def pretty_print_conversation(messages):
    for message in messages:
        color = role_to_color[message["role"]]
        m = f"{message['role']}: {message['content']}"

        if message["role"] == "function":
            m = f"function ({message['name']}): {message['content']}"
        elif message["role"] == "assistant" and "function_call" in message:
            m = f"assistant {message['function_call']['name']}({message['function_call']['arguments']})"
        print(f"[{color}]{m}[/{color}]\n")



import os
from typing import Tuple
from urllib.parse import urlparse
import tempfile

import duckdb
from duckdb import ParserException, ProgrammingError
import httpx


def uri_validator(x):
    try:
        result = urlparse(x)
        return all([result.scheme, result.netloc])
    except:
        return False


def create_duckdb(duckdb_path: str = ":memory:") -> duckdb.DuckDBPyConnection:
    # By default, duckdb is fully in-memory - we can provide a path to get
    # persistent storage

    duckdb_connection = duckdb.connect(duckdb_path)
    try:
        duckdb_connection.sql("INSTALL httpfs;")
        duckdb_connection.sql("LOAD httpfs;")
    except Exception:
        print(
            "Failed to install httpfs extension. Loading remote files will not be supported"
        )

    duckdb_connection.sql(
        "create table if not exists qabot_queries(query VARCHAR, timestamp TIMESTAMP DEFAULT CURRENT_TIMESTAMP);"
    )

    return duckdb_connection


def import_into_duckdb_from_files(
        duckdb_connection: duckdb.DuckDBPyConnection,
        files: list[str],
        dangerously_allow_write_access=False
) -> Tuple[duckdb.DuckDBPyConnection, list[str]]:
    executed_sql = []
    for i, file_path in enumerate(files, 1):
        if file_path.startswith("postgresql://"):
            try:
                duckdb_connection.sql("INSTALL postgres_scanner;")
            except Exception:
                print(
                    "Failed to install postgres_scanner extension. Loading directly from postgresql will not be supported"
                )
                continue
            db_type = "(TYPE postgres, READ_ONLY)" if not dangerously_allow_write_access else "(TYPE postgres)"
            duckdb_connection.execute(f"ATTACH '{file_path}' as postgres_db {db_type};")

            _set_search_path(duckdb_connection)
        elif file_path.endswith('.json'):
            # use the filename as the table name
            table_name, _ = os.path.splitext(os.path.basename(file_path))
            # remove any non-alphanumeric characters
            new_table_name = "".join([c for c in table_name if c.isalnum()])
            duckdb_connection.execute(f"CREATE TABLE {new_table_name} AS select * from read_json('{file_path}');")
            _set_search_path(duckdb_connection)
        elif file_path.endswith('.xlsx'):
            try:
                duckdb_connection.sql("INSTALL spatial;")
                duckdb_connection.sql("LOAD spatial;")
            except Exception:
                print(
                    "Failed to install spatial extension. Loading directly from Excel files will not be supported"
                )
                continue
            # use the filename as the table name
            table_name, _ = os.path.splitext(os.path.basename(file_path))
            # remove any non-alphanumeric characters
            new_table_name = "".join([c for c in table_name if c.isalnum()])
            duckdb_connection.execute(f"CREATE TABLE {new_table_name} AS select * from st_read('{file_path}');")
            _set_search_path(duckdb_connection)
        elif file_path.endswith(".sqlite"):
            try:
                duckdb_connection.sql("INSTALL sqlite;")
                duckdb_connection.sql("LOAD sqlite;")
            except Exception:
                print(
                    "Failed to install sqlite extension. Loading directly from sqlite will not be supported"
                )
                continue
            #duckdb_connection.execute(f"CALL sqlite_attach('{file_path}')")
            query = f"ATTACH '{file_path}' as sqlite_db (TYPE SQLITE);"
            duckdb_connection.execute(query)
            _set_search_path(duckdb_connection)

            executed_sql.append(query)
        else:
            executed_sql.append(
                load_external_data_into_db(duckdb_connection, file_path)
            )

    return duckdb_connection, executed_sql


def _set_search_path(duckdb_connection: duckdb.DuckDBPyConnection):
    db_names = [x[0] for x in duckdb_connection.sql(f"SELECT database_name FROM duckdb_databases() where internal = false;").fetchall()]
    query = f"set search_path = '{','.join(db_names)}';"
    duckdb_connection.execute(query)

def load_external_data_into_db(
    conn: duckdb.DuckDBPyConnection, file_path, allow_view=True
):
    # Work out if the filepath is actually a url (e.g. s3://)
    is_url = uri_validator(file_path)
    # Get the file name without extension from the file_path
    table_name, extension = os.path.splitext(os.path.basename(file_path))
    # If the table_name isn't a valid SQL identifier, we'll need to use something else

    table_name = (
        table_name.replace("-", "_")
        .replace(".", "_")
        .replace(" ", "_")
        .replace("/", "_")
    )

    try:
        conn.sql(f"create table t_{table_name} as select 1;")
        conn.sql(f"drop table t_{table_name};")
    except (ParserException, ProgrammingError):
        table_name = "data"

    # try to create a view then fallback to a table if it fails
    use_view = allow_view and is_url
    try:
        create_statement = f"create {'view' if use_view else 'table'} '{table_name}' as select * from '{file_path}';"
        conn.sql(create_statement)
    except duckdb.IOException:
        # This can occur if the server doesn't send Content-Length headers
        # We can work around this by downloading the data locally and then
        # loading it from there. We assume CSV files for now
        if is_url:
            with tempfile.NamedTemporaryFile(
                mode="w+b", suffix=".csv", delete=False
            ) as f:
                with httpx.stream("GET", file_path) as r:
                    for chunk in r.iter_bytes():
                        if chunk:
                            f.write(chunk)
                    f.flush()

                create_statement = f"create table '{table_name}' as select * from read_csv_auto('{f.name}');"

        conn.sql(create_statement)

    return create_statement



system_prompt = """You are Qabot, a large language model trained to helpfully answer questions involving 
data and code.

Qabot is designed to be able to assist with a wide range of tasks, from answering simple questions to 
providing in-depth explorations on a wide range of topics. Your tools run in the user's
environment so you can take advantage of their installed version of DuckDB and any data they have loaded.

Qabot prefers to answer questions by first querying information to guide its answer. Qabot responds with clarifying
questions if the request isn't clear.

Qabot prefers to split questions into small discrete steps, communicating the plan to the user at each step.
Except for straight forward requests `research` before starting.

Qabot has great SQL skills and access to a powerful DuckDB database. Qabot can and should create new tables
and views for storing intermediate results.

Unless the user specifies in their question a specific number of examples to obtain, limit 
select queries to 20 results.

Pay attention to use only the column names that you can see in the schema description. Pay attention
to which column is in which table.
    
DuckDB functions that may be helpful:
- SELECT * FROM glob('*/'); -- List of all files and folders in current directory
- SELECT size, parse_path(filename), content FROM read_text('*.md') 
- SELECT format('I''d rather be {1} than {0}.', 'right', 'happy'); -- I'd rather be happy than right.
- SELECT list_transform([1, 2, NULL, 3], x -> x + 1); -- [2, 3, NULL, 4]
- SELECT list_transform([5, NULL, 6], x -> coalesce(x, 0) + 1); -- [6, 1, 7]
- SELECT function_name,function_type,description,return_type,parameters,parameter_types FROM duckdb_functions()
- SELECT * FROM 'https://domain.tld/file.extension';


DuckDB Extensions can be used although you have to instantiate them: INSTALL spatial; LOAD spatial;
Then query using the extensions functions e.g. ST_Area, ST_Distance, ST_Point etc. Example:
```
CREATE TABLE t1 AS SELECT point::GEOMETRY AS geom
FROM st_generatepoints({min_x: 0, min_y: 0, max_x: 100, max_y: 100}::BOX_2D, 10_000, 1337);
-- Create an index on the table.
CREATE INDEX my_idx ON t1 USING RTREE (geom);
-- Perform a query with a "spatial predicate" on the indexed geometry column
-- Note how the second argument in this case, the ST_MakeEnvelope call is a "constant"
SELECT count(*) FROM t1 WHERE ST_Within(geom, ST_MakeEnvelope(45, 45, 65, 65));
```

Remember JSON is well supported by DuckDB queries and is similar to Postgresql. 

If the question does not seem related to the currently loaded data, Qabot considers other sources for 
and presents options to the user. Any queries that qabot considers malicious simply returns 
"I can't help with that" as the answer.

All interactions with the user are carried out through tool calls, Qabot's non-tool call replies are
treated as Qabot's internal monologue and should not be used as a response to the user. Instead this is 
used for summarizing failed attempts and planning out future steps.

Qabot prefers to give factual answers backed by data, calculations are to be computed by executing SQL. 
"""


research_prompt = f"""
You are the research and planning component of Qabot, a large language model based application that helpfully answers
questions involving data and code.

Qabot is designed to be able to assist with a wide range of tasks, from answering simple questions to 
providing in-depth explorations on a wide range of topics. Qabot tools run in the user's
environment. Qabot has great SQL skills and access to a powerful DuckDB database - guide Qabot to use this.

The main Qabot agent is a smaller LLM that is requesting your help formulating a plan. Ideally provide code
samples using DuckDB/Postgresql SQL. 

The system prompt for the weaker model delimited by --- --- --- is:

{system_prompt}

--- --- ---

The last messages between the user and qabot:
"""


import textwrap


def get_function_specifications(allow_wikidata: bool = True, allow_research: bool = True):
    function_specifications = [
        {
            "name": "execute_sql",
            "description": """Run SQL queries with a local DuckDB database engine. Use for accessing data, COPYing to/from files or any math computation. 
            For best results, pass in the fully qualified name: 'table_catalog.table_schema.table_name'
            
            
            """,
            "parameters": {
                "type": "object",
                "properties": {
                    "query": {
                        "type": "string",
                        "description": "DuckDB dialect SQL query. Check the table exists first.",
                    },
                },
                "required": ["query"],
            },
        },
        {
            "name": "show_tables",
            "description": "Show the locally available database tables and views",
            "parameters": {
                "type": "object",
                "properties": {},
            },
        },
        {
            "name": "describe_table",
            "description": "Show the column names and types of a local database table or view.",
            "parameters": {
                "type": "object",
                "properties": {
                    "catalog": {
                        "type": "string",
                        "description": "The catalog if known e.g. 'postgres_db'.",
                    },
                    "schema": {
                        "type": "string",
                        "description": "The schema (if known)",
                    },
                    "table": {
                        "type": "string",
                        "description": "The table or view name",
                    },
                },
                "required": ["table"],
            },
        },
        {
            "name": "load_data",
            "description": "Load data from one or more local or remote files into the local DuckDB database",
            "parameters": {
                "type": "object",
                "properties": {
                    "files": {
                        "type": "array",
                        "description": "File path or urls",
                        "items": {
                            "type": "string",
                            "examples": [
                                "data/chinook.sqlite",
                                "https://duckdb.org/data/prices.csv",
                            ],
                        },
                    },
                },
                "required": ["file"],
            },
        },
        # A special function to call to summarize the answer
        {
            "name": "answer",
            "description": "Reply to the user question with a detailed answer",
            "parameters": {
                "type": "object",
                "properties": {
                    "summary": {
                        "type": "string",
                        "description": "A standalone one-line summary answering the users question or briefly explaining why the question can't be answered",
                    },
                    "detail": {
                        "type": "string",
                        "description": """detailed answer to the user's question including how it was computed. 
                        Markdown is acceptable including code snippets, ascii graphs diagrams.""",
                    },
                    "query": {
                        "type": "string",
                        "description": "If the query is reproducible, include it here. Don't include the full schema",
                    },
                    # "value": {
                    #     "type": "any",
                    #     "description": "If the answer is a number or array, include the value here",
                    # }
                },
                "required": ["summary", "detail"],
            },
        },
    ]

    if allow_wikidata:
        function_specifications.append(
            {
                "name": "wikidata",
                "description": textwrap.dedent(
                    """Source data from Wikidata if not available locally.
                    Input to this tool is a single SPARQL statement for Wikidata. Limit all requests to 50 or fewer rows. 

                    Output is the raw response in json. If the query is not correct, an error message will be returned. 
                    If an error is returned, you may rewrite the query and try again. If you are unsure about the response
                    you can try rewrite the query and try again. Prefer local data before using this tool.
                    """
                ),
                "parameters": {
                    "type": "object",
                    "properties": {
                        "query": {
                            "type": "string",
                            "description": "A valid SPARQL query for Wikidata",
                        },
                    },
                },
            }
        )


    if allow_research:
        function_specifications.append(
            {
                "name": "research",
                "description": textwrap.dedent(
                    """Planning review, or outsource considering a question regarding how to solve a problem in depth.
                    Consult an expert! Provide full context and details.
                    
                    Example: What is an efficient way to implement Monte Carlo simulations in DuckDB for computing PI with 100 million points?
                    """
                ),
                "parameters": {
                    "type": "object",
                    "properties": {
                        "query": {
                            "type": "string",
                            "description": "A plan to consider, or another question that Qabot would like assistance with",
                        },
                    },
                },
            }
        )

    return function_specifications


from typing import List, Optional
import warnings
from openai import OpenAI
import typer
from rich import print
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.prompt import Confirm, Prompt
import httpx

from qabot.config import Settings
from qabot.functions.data_loader import import_into_duckdb_from_files, create_duckdb
from qabot.agent import Agent
from qabot.formatting import (
    format_duck,
    format_robot,
    format_user,
    format_rocket,
    ROBOT_COLOR,
    format_query,
)
from qabot.prompts import INITIAL_NON_INTERACTIVE_PROMPT, FOLLOW_UP_PROMPT

warnings.filterwarnings("ignore")

app = typer.Typer(pretty_exceptions_show_locals=False, pretty_exceptions_enable=False)


def handle_db(agent, arg: str):
    try:
        result = agent.functions["execute_sql"](query=arg)
        print(result)
    except Exception as e:
        print(f"[red]Error executing SQL: {e}[/red]")

def handle_describe(agent, arg: str):
    try:
        result = agent.functions["describe_table"](table=arg)
        print(format_duck(result))
    except Exception as e:
        print(f"[red]Error describing table: {e}[/red]")

def handle_help(agent, arg: str):
    print("Available commands:")
    print("  /db <SQL>         Execute SQL directly on DuckDB")
    print("  /help             Show this help message")
    print("  /exit             Exit the CLI")
    print("Anything else is sent to the LLM")


# Create a command registry
COMMAND_HANDLERS = {
    "db": handle_db,
    "help": handle_help,
    "exit": lambda agent, arg: exit(0),
}

@app.command()
def main(
    query: str = typer.Option(
        "Describe the tables", "-q", "--query", prompt=INITIAL_NON_INTERACTIVE_PROMPT
    ),
    file: Optional[List[str]] = typer.Option(
        None, "-f", "--file", help="File or url containing data to load and query"
    ),
    prompt_context: Optional[str] = typer.Option(
        None, "--context", help="File or url containing text data to use in the LLM prompt - e.g. describing the data"
    ),
    database_uri: Optional[str] = typer.Option(
        ":memory:",
        "-d",
        "--database",
        help="DuckDB Database URI (e.g. '/tmp/qabot.duckdb')",
    ),
    enable_wikidata: bool = typer.Option(
        False, "-w", "--wikidata", help="Allow querying from wikidata"
    ),
    verbose: bool = typer.Option(
        False, "-v", "--verbose", help="Essentially debug output"
    ),
):
    """
    Query a database or Wikidata using a simple natural language query.

    Example:
        qabot -q "What is the average length of Queen songs?" -f data/Chinook.sqlite
    """

    settings = Settings()
    executed_sql = ""
    # If files are given load data into local DuckDB
    print(format_duck("Creating local DuckDB database..."))
    if enable_wikidata:
        print(format_duck("Enabling Wikidata..."))
    database_engine = create_duckdb(database_uri)

    openai_client = OpenAI(
        api_key=settings.OPENAI_API_KEY,
        base_url=settings.OPENAI_BASE_URL,
    )

    if file and len(file) > 0:
        if isinstance(file, str):
            file = [file]
        print(format_duck("Loading data..."))
        database_engine, executed_sql = import_into_duckdb_from_files(
            database_engine, file
        )
        executed_sql = "\n".join(executed_sql)
        print(format_query(executed_sql))

    # Load the optional context data
    context_data = ""
    if prompt_context is not None:
        try:
            if prompt_context.startswith("http://") or prompt_context.startswith("https://"):
                response = httpx.get(prompt_context)
                response.raise_for_status()  # Raises an HTTPStatusError if the response status code is 4XX/5XX
                context_data = response.text
            else:
                with open(prompt_context, 'r', encoding='utf-8') as file:
                    context_data = file.read()
        except Exception as e:
            raise RuntimeError(f"Failed to load context data from {prompt_context}: {e}")

    
    with Progress(
        SpinnerColumn(),
        TextColumn("[green][progress.description]{task.description}"),
        transient=False,
    ) as progress:
        t2 = progress.add_task(
            description=format_rocket("Sending query to LLM"), total=None
        )

        def clarification(clarification):
            progress.stop()
            return Prompt.ask(
                format_rocket("Clarification requested:\n")
                + format_robot(clarification)
            )

        def terminate_session(message: str):
            progress.stop()
            print(format_robot(message))
            raise SystemExit

        agent = Agent(
            database_engine=database_engine,
            verbose=verbose,
            models=settings.agent_model,
            allow_wikidata=settings.QABOT_ENABLE_WIKIDATA and enable_wikidata,
            terminate_session_callback=terminate_session,
            clarification_callback=clarification
            if settings.QABOT_ENABLE_HUMAN_CLARIFICATION
            else None,
            prompt_context=context_data,
            openai_client=openai_client,
        )

        progress.remove_task(t2)

        while True:
            # Check if the input is a command (starts with "/")
            if query.startswith('/'):
                # Split command and arguments
                parts = query.strip().split(maxsplit=1)
                cmd = parts[0][1:]  # remove the leading '/'
                arg = parts[1] if len(parts) > 1 else ''

                handler = COMMAND_HANDLERS.get(cmd)
                # Stop progress to ensure output displays correctly
                progress.stop()
                print()
                if handler:
                    handler(agent, arg)
                else:
                    print(f"[red]Unknown command: {cmd}[/red]")

                print()
                # Prompt for next input after a command and continue to next loop iteration
                query = Prompt.ask(FOLLOW_UP_PROMPT)
                if query.lower() in {'n', 'no', 'q', 'exit', 'quit'}:
                    break
                #progress.start()
                continue

            print(format_rocket(f"Sending query to LLM ({settings.agent_model.default_model_name})"))
            print(format_user(query))

            t = progress.add_task(description="Processing query...", total=None)
            result = agent(query)

            # Stop the progress before outputting result and prompting for any more input
            progress.remove_task(t)
            progress.stop()
            print()


            if verbose:
                # Likely the users query was quite a ways back in the console history
                print(format_rocket("Question:"))
                print(format_user(query))

            if result:
                print(format_robot(result["summary"]))
                print()
                if "detail" in result:
                    print(f"[{ROBOT_COLOR}]\n{result['detail']}\n")

                if "query" in result:
                    print(format_query(result["query"]))

            print()
            query = Prompt.ask(FOLLOW_UP_PROMPT)

            if query.lower() in {'n', 'no', "q", "exit", "quit"}:
                #  and Confirm.ask(
                #                 "Are you sure you want to Quit?"
                #             )
                break

            progress.start()


def run():
    app()


if __name__ == "__main__":
    run()
